/* Put in the public domain 2000 by Sam Trenholme */
/* Kiwi Parse: A series of programs that parse the .kiwirc file and set the
   relevent global variables */

#include "JessStr/JessStr.h"

/* Alphanumberic digits */
#define AN "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890-_"

/* Keywords that are significant to Kiwi's rc file */

#define KEYCOUNT 14

char *keywords[KEYCOUNT] = { 
	"kiwi_qmail", 
	"kiwi_sendmail_path", 
	"kiwi_days_short",
	"kiwi_days_mid",
	"kiwi_days_long",
	"kiwi_forward",
	"kiwi_append",
	"kiwi_my_mailbox",
	"kiwi_my_address",
	"kiwi_maillog",
	"kiwi_key",
	"kiwi_password",
	"kiwi_mid_process",
        "kiwi_bouncemail" };

/* kiwi_goodjs: Determine if a given js_string object is a valid string
                for use in the Kiwi internals
   input: Pointer to js_string object to test
   ouput: JS_ERROR if bad, JS_SUCCESS if good */
int kiwi_goodjs(js_string *test) {
    if(js_has_sanity(test) == JS_ERROR)
        return JS_ERROR;
    return JS_SUCCESS;
    }

/* keyword2num: Convert a keyword (like "kiwi_maillog") to a number
                (10, in this case)
   input: A js_string object with the keyword
   output: The number of the keyword (starting at 1), JS_ERROR on error,
           -2 on no match */
int keyword2num(js_string *keyword) {
    int counter = 0;
    js_string *name;

    if(kiwi_goodjs(keyword) == JS_ERROR)
        return JS_ERROR;

    if((name = js_create(256,1)) == 0)
        return JS_ERROR;

    while(counter<KEYCOUNT) {
        if(js_qstr2js(name,keywords[counter]) == JS_ERROR);
	    return JS_ERROR;
	if(js_issame(keyword,name))
	    return counter;
        counter++;
        }
    
    return -2;
    }

/* num2keyword: convert a number in to a keyword
   input: A number to make the keyword, the place to store the keyword
   output: JS_ERROR if it is out of range or any other error, JS_SUCCESS
           otherwise */
int num2keyword(int num, js_string *keyword) {
    if(kiwi_goodjs(keyword) == JS_ERROR)
        return JS_ERROR;

    if(num < 0 || num >=KEYCOUNT)
        return JS_ERROR;

    return js_qstr2js(keyword,keywords[num]);
    }

/* Parseline: Given a line of the Kiwirc file, parse that line
   input: A js_String object pointing to the contents of the line,
          A js_string object that will contain the name of the variable
          on the line ("ERROR" on syntax error),
          A js_string object that will contain the variable's value
          (description of problem on syntax error)
   output: JS_ERROR on fatal error, JS_SUCCESS on success */
int parseline(js_string *line, js_string *var, js_string *value) {

    js_string *quotes, *alphanumeric, *equals, *hashq, *bslashq;
    int quote1, quote2; /* Location of quotes in js_string line */
    int varstart,varend, valstart, valend; /* Pointers to beginning and end
                                              of the variable name and the
                                              value for the variable */
    int equalp, hashp; /* Location on equals sign and of hash */
    char quote = '"';
    char equal = '=';
    char hash = '#';
    char bslash = '\\';

    /* Sanity checks */
    if(kiwi_goodjs(line) == JS_ERROR)
        return JS_ERROR;
    if(kiwi_goodjs(var) == JS_ERROR)
        return JS_ERROR;
    if(kiwi_goodjs(value) == JS_ERROR)
        return JS_ERROR;
    /* Alocate strings */
    if((quotes = js_create(256,1)) == 0)
        return JS_ERROR;
    if((alphanumeric = js_create(256,1)) == 0)
        return JS_ERROR;
    if((equals = js_create(256,1)) == 0)
        return JS_ERROR;
    if((hashq = js_create(256,1)) == 0)
        return JS_ERROR;
    if((bslashq = js_create(256,1)) == 0)
        return JS_ERROR;

    /* Initialize the various sets we look for */
    js_str2js(quotes,&quote,1,1); 
    js_str2js(equals,&equal,1,1); 
    js_str2js(hashq,&hash,1,1); 
    js_str2js(bslashq,&bslash,1,1); 
    js_qstr2js(alphanumeric,AN);

    /* By defulat, no quotes on the line */ 
    quote1 = quote2 = -2;    

    /* Locate the quotes */
    if((quote1 = js_match(quotes,line)) == JS_ERROR)
        return JS_ERROR;
    if(quote1 != -2)
        quote2 = js_match_offset(quotes,line,quote1 + 1);
   
    /* Locate the hash */
    if((hashp = js_match(hashq,line)) == JS_ERROR)
        return JS_ERROR;
    /* Take it beyond the end of the line if not found or in quotes */
    if(hashp == -2 || (hashp > quote1 && hashp < quote2))
        hashp = js_length(line) + 1;
    /* If the quotes are after the hash, remove them */
    if(quote1 > hashp)
        quote1 = quote2 = -2;

    /* Locate the equals */
    if((equalp = js_match(equals,line)) == JS_ERROR)
        return JS_ERROR;

    /* Find the beginning and end of the first word */
    varstart = varend = -2;
    if((varstart = js_match(alphanumeric,line)) == JS_ERROR)
        return JS_ERROR; 
    if(varstart != -2 && varstart < hashp) {
        varend = varstart;
        while(js_match_offset(alphanumeric,line,varend + 1) == varend + 1)
	    varend++;
        }
    else if(varstart >= hashp)
        varstart = -2;
    /* Doesn't count if you are in quotes -- this is a syntax error */
    if(varstart > quote1 && varend < quote2) {
        js_qstr2js(var,"ERROR");
        js_qstr2js(value,"Variable name in quotes");
        return JS_SUCCESS;
        }

    /* find the beginning and end of the second word */
    valstart = valend = -2;
    if(quote1 != -2 && quote2 != -2) {
        valstart = quote1;
	valend = quote2;
	}
    else {
        if((valend = js_match_offset(alphanumeric,line,varend + 1)) == 
	   JS_ERROR)
	    return JS_ERROR;
        if(valend != -2 && valend < hashp) {
	    valend = valstart;
	    while(js_match_offset(alphanumeric,line,valend + 1) == valend + 1)
	        valend++;
            }
        else if(valstart >= hashp)
	    valstart = valend = -2;
        }

    /* return error if there is backslash in line */
    if(js_match(bslashq,line) != -2) {
        js_qstr2js(var,"ERROR");
        js_qstr2js(value,"Backslash not supported yet");
        return JS_SUCCESS;
        }
        
    /* Now that va[rl]start and va[rl]end have values, do the substrs */

    if(varstart >=0 && varend >= varstart) {
        /* Syntax error if before equals sign */
        if(varstart > equalp) {
            js_qstr2js(var,"ERROR");
            js_qstr2js(value,"Equal sign before variable name");
            return JS_SUCCESS;
            }
        if((js_substr(line,var,varstart,varend - varstart + 1)) == JS_ERROR)
            return JS_ERROR;
        }
    
    if(valstart >=0 && valend >= valstart) {
        /* Syntax error if after equals sign */
        if(valstart < equalp) {
            js_qstr2js(var,"ERROR");
            js_qstr2js(value,"Equal sign after variable value");
            return JS_SUCCESS;
            }
        if((js_substr(line,value,valstart,valend - valstart + 1)) == JS_ERROR)
            return JS_ERROR;
        }

    return JS_SUCCESS;

    }

